#include <Python.h>
#include <gssapi.h>
#include <gssapi/gssapi_krb5.h>
#include <stdio.h>

#include "py_compat.h"

#define DEBUG(str, args...) printf("  %s: " str "\n", __func__, args)

struct module_state {
    PyObject *GSSError_class;
    PyObject *RequirementFlag_class;
    PyObject *MechType_class;
    PyObject *NameType_class;
};

STATESTUB;

#ifdef PY3K
    static int
    impl_traverse(PyObject *m, visitproc visit, void *arg)
    {
        Py_VISIT(GETSTATE(m)->GSSError_class);
        Py_VISIT(GETSTATE(m)->RequirementFlag_class);
        Py_VISIT(GETSTATE(m)->MechType_class);
        return 0;
    }

    static int
    impl_clear(PyObject *m)
    {
        Py_DECREF(GETSTATE(m)->GSSError_class);
        Py_DECREF(GETSTATE(m)->RequirementFlag_class);
        Py_DECREF(GETSTATE(m)->MechType_class);
        return 0;
    }
#endif

/* Utility method to raise an error */
static void raise_gss_error(PyObject *m, OM_uint32 maj, OM_uint32 min)
{
    PyErr_SetObject(GETSTATE(m)->GSSError_class,
                    Py_BuildValue("II", maj, min));
}

static PyObject *
importName(PyObject *self, PyObject *args)
{
    const char *name;
    int name_len;
    int raw_name_type = 0; /* default: hostbased_service */

    if(!PyArg_ParseTuple(args, "s#|i", &name, &name_len, &raw_name_type))
        return NULL;

    gss_OID name_type;
    switch(raw_name_type) {
        case 0:
            name_type = GSS_C_NT_HOSTBASED_SERVICE;
            break;
        case 1:
            name_type = GSS_KRB5_NT_PRINCIPAL_NAME;
            break;
        case 2:
            name_type = GSS_C_NT_USER_NAME;
            break;
        case 3:
            name_type = GSS_C_NT_ANONYMOUS;
            break;
        case 4:
            name_type = GSS_C_NT_MACHINE_UID_NAME;
            break;
        case 5:
            name_type = GSS_C_NT_STRING_UID_NAME;
            break;
        case 6:
            name_type = GSS_C_NT_EXPORT_NAME;
            break;
        default:
            name_type = NULL;
            /* TODO(sross): throw error */
            break;
    }

    gss_buffer_desc name_token = GSS_C_EMPTY_BUFFER;
    name_token.length = name_len;
    name_token.value = name;

    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    gss_name_t output_name;

    Py_BEGIN_ALLOW_THREADS
        maj_stat = gss_import_name(&min_stat, &name_token,
                                   name_type, &output_name);
    Py_END_ALLOW_THREADS

    if (maj_stat == GSS_S_COMPLETE) {
        PyObject *out_name_obj = PyCapsule_New(output_name, NULL, NULL);
        return out_name_obj;
    }
    else {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }
}
#define COMPARE_OIDS(a,b) ( (a->length == b->length) && \
                            !memcmp(a->elements, b->elements, a->length) )

#define CHECK_AND_RETURN_NAME_TYPE(var, target, pytarget) \
    if (COMPARE_OIDS(var, target)) { \
        PyObject *argList = Py_BuildValue("(I)", pytarget); \
        PyObject *res = PyObject_CallObject(GETSTATE(self)->NameType_class, \
                                            argList); \
        Py_DECREF(argList); \
        return res; \
    }

static PyObject *
createNameType(PyObject *self, gss_OID name_type)
{
    CHECK_AND_RETURN_NAME_TYPE(name_type, GSS_C_NT_HOSTBASED_SERVICE, 0);
    CHECK_AND_RETURN_NAME_TYPE(name_type, GSS_KRB5_NT_PRINCIPAL_NAME, 1);
    CHECK_AND_RETURN_NAME_TYPE(name_type, GSS_C_NT_USER_NAME, 2);
    CHECK_AND_RETURN_NAME_TYPE(name_type, GSS_C_NT_ANONYMOUS, 3);
    CHECK_AND_RETURN_NAME_TYPE(name_type, GSS_C_NT_MACHINE_UID_NAME, 4);
    CHECK_AND_RETURN_NAME_TYPE(name_type, GSS_C_NT_STRING_UID_NAME, 5);
    CHECK_AND_RETURN_NAME_TYPE(name_type, GSS_C_NT_EXPORT_NAME, 6);

    /* if we haven't returned something else already... */
    Py_RETURN_NONE;
}

#define GET_CAPSULE(type, obj) (type)PyCapsule_GetPointer(obj, NULL)
#define GET_CAPSULE_DEREF(type, obj) *((type*)PyCapsule_GetPointer(obj, NULL))

static PyObject *
displayName(PyObject *self, PyObject *raw_name)
{
    gss_name_t name = GET_CAPSULE(gss_name_t, raw_name);

    gss_buffer_desc name_buffer = GSS_C_EMPTY_BUFFER;
    gss_OID name_type;

    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    Py_BEGIN_ALLOW_THREADS
        maj_stat = gss_display_name(&min_stat, name, &name_buffer,
                                    &name_type);
    Py_END_ALLOW_THREADS

    if (maj_stat == GSS_S_COMPLETE) {
        PyObject *python_name_type = createNameType(self, name_type);
        PyObject *res = Py_BuildValue(BYTE_TUPLE_STR("", "O"),
                                      name_buffer.value,
                                      name_buffer.length,
                                      python_name_type);
        Py_DECREF(python_name_type);
        return res;
    }
    else {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }
}

static PyObject *
compareName(PyObject *self, PyObject *args)
{
    PyObject *raw_name1;
    PyObject *raw_name2;

    if (!PyArg_ParseTuple(args, "OO", &raw_name1, &raw_name2))
        return NULL;

    gss_name_t name1 = GET_CAPSULE(gss_name_t, raw_name1);
    gss_name_t name2 = GET_CAPSULE(gss_name_t, raw_name2);

    int name_eq;

    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    maj_stat = gss_compare_name (&min_stat, name1, name2, &name_eq);

    if (maj_stat == GSS_S_COMPLETE) {
        if (name_eq) {
            Py_INCREF(Py_True);
            return Py_True;
        }
        else {
            Py_INCREF(Py_False);
            return Py_False;
        }
    }
    else {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }
}

static PyObject *
exportName(PyObject *self, PyObject *raw_name)
{
    gss_name_t name = GET_CAPSULE(gss_name_t, raw_name);

    gss_buffer_desc exported_name = GSS_C_EMPTY_BUFFER;

    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    maj_stat = gss_export_name(&min_stat, name, &exported_name);

    if (maj_stat == GSS_S_COMPLETE) {
        return Py_BuildValue(BYTE_STR, exported_name.value,
                             exported_name.length);
    }
    else {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }
}

static PyObject *
releaseName(PyObject *self, PyObject *raw_name)
{
    gss_name_t name = GET_CAPSULE(gss_name_t, raw_name);

    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    maj_stat = gss_release_name(&min_stat, &name);

    if (maj_stat == GSS_S_COMPLETE) {
        Py_RETURN_NONE;
    }
    else {
       raise_gss_error(self, maj_stat, min_stat);
       return NULL;
    }
}

static PyObject *
createMechType(PyObject *self, gss_OID mech_type)
{
    if (COMPARE_OIDS(mech_type, gss_mech_krb5)) {
        /* return MechType.kerberos */
        PyObject *argList = Py_BuildValue("(I)", 0);
        PyObject *res = PyObject_CallObject(GETSTATE(self)->MechType_class,
                                            argList);
        Py_DECREF(argList);
        return res;
    }
    else {
        Py_RETURN_NONE;
    }
}

static PyObject *
createMechList(PyObject *self, gss_OID_set mechs)
{
    PyObject *list = PyList_New(0);
    int i;
    for (i = 0; i < mechs->count; i++) {
        PyObject *mech = createMechType(self, &(mechs->elements[i]));

        if (mech != Py_None) {
            PyList_Append(list, mech);
        }

        Py_DECREF(Py_None);
    }

    return list;
}

#define INT_CHECK_AND_ADD_TO_SET(list, pval, cval, set) \
    do { \
        PyObject *num = PyInt_FromLong(pval); \
        CHECK_AND_ADD_TO_SET(list, pval, cval, set); \
        Py_DECREF(num); \
    } while(0);

#define CHECK_AND_ADD_TO_SET(list, pval, cval, set) \
    if (PySequence_Contains(list, pval)) { \
        OM_uint32 min_stat_add; \
        gss_add_oid_set_member(&min_stat_add, cval, &set); \
    }

static gss_OID_set
createOIDMechSet(PyObject *list)
{
    OM_uint32 min_stat_create;
    gss_OID_set set;
    gss_create_empty_oid_set(&min_stat_create, &set);

    /* MechType.kerberos = 0 */
    INT_CHECK_AND_ADD_TO_SET(list, 0, gss_mech_krb5, set);

    return set;
}

static PyObject *
indicateMechs(PyObject *self, PyObject *noargs)
{
    gss_OID_set mech_set;

    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    maj_stat = gss_indicate_mechs(&min_stat, &mech_set);

    if (maj_stat == GSS_S_COMPLETE)
    {
        PyObject *mech_list = createMechList(self, mech_set);
        OM_uint32 min_stat_release;
        gss_release_oid_set(&min_stat_release, &mech_set);

        return mech_list;
    }
    else
    {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }
}

#define CAPSULE_OR_DEFAULT(type, obj, def) \
            ( obj == Py_None ? def : GET_CAPSULE(type, obj))

#define CAP_OR_DFLT_DR(type, obj, def) \
            ( obj == Py_None ? def : GET_CAPSULE_DEREF(type, obj))

#define VALUE_OR_DEFAULT(obj, val, def) \
            ( obj == Py_None ? def : val )

#define TRUE_FALSE_NONE(obj, t, f, n) \
            ( obj == Py_None ? n : ( PyLong_AsLong(obj) ? t : f ))

static PyObject *
acquireCred(PyObject *self, PyObject *args, PyObject *keywds)
{
    PyObject *raw_desired_name; /* capsule */
    OM_uint32 input_ttl = 0; /* default: 0 */
    PyObject *raw_mechs = Py_None; /* capsule or default: None */
    PyObject *raw_cred_usage = Py_None; /* boolean or default: None */

    static char *kwlist[]  = {"name", "ttl", "mechs", "cred_usage", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "O|IOO", kwlist,
                                     &raw_desired_name, &input_ttl,
                                     &raw_mechs, &raw_cred_usage))
        return NULL;

    gss_name_t desired_name = GET_CAPSULE(gss_name_t, raw_desired_name);

    gss_OID_set desired_mechs = VALUE_OR_DEFAULT(raw_mechs,
                                                 createOIDMechSet(raw_mechs),
                                                 GSS_C_NO_OID_SET);

    gss_cred_usage_t cred_usage = TRUE_FALSE_NONE(raw_cred_usage,
                                                  GSS_C_ACCEPT,
                                                  GSS_C_INITIATE,
                                                  GSS_C_BOTH);

    gss_cred_id_t creds;
    gss_OID_set actual_mechs;
    OM_uint32 actual_ttl;

    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    Py_BEGIN_ALLOW_THREADS
        maj_stat = gss_acquire_cred(&min_stat, desired_name, input_ttl,
                                    desired_mechs, cred_usage, &creds,
                                    &actual_mechs, &actual_ttl);
    Py_END_ALLOW_THREADS


    if (raw_mechs != Py_None) {
        OM_uint32 min_stat_release;
        gss_release_oid_set(&min_stat_release, &desired_mechs);
    }

    if (maj_stat == GSS_S_COMPLETE) {
        PyObject *cap_creds = PyCapsule_New(creds, NULL, NULL);
        PyObject *list_mechs = createMechList(self, actual_mechs);
        PyObject *res = Py_BuildValue("OOI",
                                      cap_creds, list_mechs, actual_ttl);
        Py_DECREF(cap_creds);
        Py_DECREF(list_mechs);
        return res;
    }
    else {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }
}

static PyObject *
releaseCred(PyObject *self, PyObject *raw_creds)
{
    gss_cred_id_t creds = GET_CAPSULE(gss_cred_id_t, raw_creds);

    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    maj_stat = gss_release_cred(&min_stat, &creds);

    if (maj_stat == GSS_S_COMPLETE) {
        Py_RETURN_NONE;
    }
    else {
       raise_gss_error(self, maj_stat, min_stat);
       return NULL;
    }
}

static PyObject *
deleteSecContext(PyObject *self, PyObject *args)
{
    PyObject *raw_context; /* capsule */
    int output_needed = 0; /* boolean, default: False */

    if(!PyArg_ParseTuple(args, "O|i", &raw_context, &output_needed))
        return NULL;

    gss_ctx_id_t ctx = GET_CAPSULE(gss_ctx_id_t, raw_context);

    gss_buffer_desc output_token = GSS_C_EMPTY_BUFFER;
    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    if (output_needed) {
        maj_stat = gss_delete_sec_context(&min_stat, &ctx, &output_token);
    }
    else {
        maj_stat = gss_delete_sec_context(&min_stat, &ctx, GSS_C_NO_BUFFER);
    }


    if (maj_stat == GSS_S_COMPLETE) {
        if (output_needed) {
            PyObject *res = Py_BuildValue(BYTE_STR,
                                          output_token.value,
                                          output_token.length);
            gss_release_buffer(&min_stat, &output_token);
            return res;
        }
        else {
            Py_RETURN_NONE;
        }
    }
    else {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }
}

static gss_OID
getMechTypeInternal(int mech_type)
{
    switch (mech_type) {
        case 0:
            return gss_mech_krb5;
        default:
            /* TODO(sross): raise exception */
            return GSS_C_NO_OID;
    }
}

static PyObject *
getMechanismType(PyObject *self, PyObject *args)
{
    int raw_mech_type; /* MechType (int) */

    if (!PyArg_ParseTuple(args, "i", &raw_mech_type))
        return NULL;

    gss_OID mech_type = getMechTypeInternal(raw_mech_type);

    PyObject *cap_mech_type = PyCapsule_New(mech_type, NULL, NULL);
    return cap_mech_type;
}

static PyObject *
canonicalizeName(PyObject *self, PyObject *args)
{
    PyObject *raw_name;
    int raw_mech_type;

    if (!PyArg_ParseTuple(args, "Oi", &raw_name, &raw_mech_type))
        return NULL;

    gss_name_t name = GET_CAPSULE(gss_name_t, raw_name);
    gss_OID mech_type = getMechTypeInternal(raw_mech_type);

    gss_name_t output_name;

    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    maj_stat = gss_canonicalize_name(&min_stat, name, mech_type, &output_name);

    if (maj_stat == GSS_S_COMPLETE) {
        return PyCapsule_New(output_name, NULL, NULL);
    }
    else {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }
}

static PyObject *
duplicateName(PyObject *self, PyObject *raw_name)
{
    gss_name_t name = GET_CAPSULE(gss_name_t, raw_name);

    gss_name_t new_name;

    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    maj_stat = gss_duplicate_name(&min_stat, name, &new_name);

    if (maj_stat == GSS_S_COMPLETE) {
        return PyCapsule_New(new_name, NULL, NULL);
    }
    else {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }
}

static int
parseFlags(PyObject *flags_list, int default_flags)
{
    if (flags_list == Py_None || flags_list == NULL) {
        return default_flags;
    }

    int cflags = 0;

    int i;
    for (i = 0; i < PyList_Size(flags_list); i++) {
        PyObject *raw_item = PyList_GetItem(flags_list, i);
        int flag = PyInt_AsLong(PyNumber_Int(raw_item));
        cflags = cflags | flag;
    }

    return cflags;
}

#define CHECK_AND_INIT_FLAG(name) \
    if (cflags & GSS_C_ ## name ## _FLAG) \
    { \
        PyObject *argList = Py_BuildValue("(I)", GSS_C_ ## name ## _FLAG); \
        PyObject *resItem = \
            PyObject_CallObject(GETSTATE(self)->RequirementFlag_class, \
                                argList); \
        PyList_Append(flag_list, resItem); \
        Py_XDECREF(resItem); \
        Py_DECREF(argList); \
    }

static PyObject *
createFlagsList(PyObject *self, int cflags)
{
    PyObject *flag_list = PyList_New(0);

    CHECK_AND_INIT_FLAG(DELEG)
    CHECK_AND_INIT_FLAG(MUTUAL)
    CHECK_AND_INIT_FLAG(REPLAY)
    CHECK_AND_INIT_FLAG(SEQUENCE)
    CHECK_AND_INIT_FLAG(CONF)
    CHECK_AND_INIT_FLAG(INTEG)
    CHECK_AND_INIT_FLAG(ANON)
    CHECK_AND_INIT_FLAG(TRANS)

    return flag_list;
}

static PyObject *
acceptSecContext(PyObject *self, PyObject *args, PyObject *keywds)
{
    char *raw_input_token = NULL; /* not null terminated, default: None/NULL */
    int raw_input_token_len;
    PyObject *raw_acceptor_cred = Py_None; /* capsule of default: None */
    PyObject *raw_ctx = Py_None; /* capsule or default: None */
    PyObject *raw_channel_bindings = Py_None; /* capsule or default: None */

    static char *kwlist[] = {"input_token", "acceptor_cred", "ctx", "channel_bindings", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "s#|OOO", kwlist,
                                     &raw_input_token, &raw_input_token_len,
                                     &raw_acceptor_cred, &raw_ctx,
                                     &raw_channel_bindings))
        return NULL;


    /* NOTE: a server SHOULD use a credential obtained by calling acquireCred or
     *       addCred for the GSS_C_NO_NAME desired_name and OID of kerberos for the
     *       mech type, but may use either GSS_C_NO_CREDENTIAL or aqquire/add cred for
     *       the server's principal name and the kerberos mechanism
     *
     *       We use the easiest way for the default for now :-)
     */
    gss_cred_id_t acceptor_cred = CAPSULE_OR_DEFAULT(gss_cred_id_t,
                                                     raw_acceptor_cred,
                                                     GSS_C_NO_CREDENTIAL);

    gss_ctx_id_t ctx = CAPSULE_OR_DEFAULT(gss_ctx_id_t, raw_ctx,
                                          GSS_C_NO_CONTEXT);

    gss_channel_bindings_t bdng = CAP_OR_DFLT_DR(gss_channel_bindings_t,
                                                 raw_channel_bindings,
                                                 GSS_C_NO_CHANNEL_BINDINGS);

    gss_buffer_desc input_token = GSS_C_EMPTY_BUFFER;

    if (raw_input_token && *raw_input_token)
    {
        input_token.length = raw_input_token_len;
        input_token.value = raw_input_token;
    }

    gss_name_t src_name;
    gss_OID mech_type; /* read-only (don't free) */
    gss_buffer_desc output_token = GSS_C_EMPTY_BUFFER;
    OM_uint32 ret_flags;
    OM_uint32 output_ttl;
    gss_cred_id_t delegated_cred;

    OM_uint32 maj_stat;
    OM_uint32 min_stat;


    Py_BEGIN_ALLOW_THREADS
        maj_stat = gss_accept_sec_context(&min_stat, &ctx, acceptor_cred,
                                          &input_token, bdng, &src_name,
                                          &mech_type, &output_token,
                                          &ret_flags, &output_ttl,
                                          &delegated_cred);
    Py_END_ALLOW_THREADS


    if (maj_stat == GSS_S_COMPLETE || maj_stat == GSS_S_CONTINUE_NEEDED) {
        PyObject *cap_ctx = PyCapsule_New(ctx, NULL, NULL);
        PyObject *cap_src_name = PyCapsule_New(src_name, NULL, NULL);
        PyObject *cap_mech_type = createMechType(self, mech_type);
        PyObject *reqs_out = createFlagsList(self, ret_flags);

        PyObject *cap_delegated_cred;
        if (delegated_cred == GSS_C_NO_CREDENTIAL) {
            cap_delegated_cred = Py_None;
            Py_INCREF(Py_None);
        }
        else {
            cap_delegated_cred = PyCapsule_New(delegated_cred, NULL, NULL);
        }

        PyObject *continue_needed;
        if (maj_stat == GSS_S_CONTINUE_NEEDED) {
            continue_needed = Py_True;
            Py_INCREF(Py_True);
        }
        else {
            continue_needed = Py_False;
            Py_INCREF(Py_False);
        }

        PyObject *res = Py_BuildValue(BYTE_TUPLE_STR("OOO","OIOO"),
                                      cap_ctx, cap_src_name,
                                      cap_mech_type, output_token.value,
                                      output_token.length, reqs_out,
                                      output_ttl, cap_delegated_cred,
                                      continue_needed);
        gss_release_buffer(&min_stat, &output_token);
        Py_DECREF(reqs_out);
        Py_DECREF(cap_ctx);
        Py_DECREF(cap_src_name);
        Py_DECREF(cap_mech_type);
        Py_DECREF(cap_delegated_cred);
        Py_DECREF(continue_needed);
        return res;
    }
    else
    {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }

}

static PyObject *
initSecContext(PyObject *self, PyObject *args, PyObject *keywds)
{
    PyObject* raw_target_name; /* capsule */
    PyObject* raw_cred = Py_None; /* capsule or default: None */
    PyObject* raw_ctx = Py_None; /* capsule or default: None */
    PyObject* raw_mech_type = Py_None; /* int or default: None */
    PyObject* flags_list = Py_None; /* list of ints, default: [MUTUAL, SEQUENCE] */
    OM_uint32 ttl = 0; /* default: 0 */
    PyObject* raw_channel_bindings = Py_None; /* capsule or default: None */
    char *raw_input_token = NULL; /* not null terminated, default: None/NuLL */
    int raw_input_token_len;

    static char *kwlist[] = {"target_name", "cred", "context", "mech_type", "flags", "time", "channel_bindings", "input_token", NULL};

    if(!PyArg_ParseTupleAndKeywords(args, keywds, "O|OOOOIOs#", kwlist,
                                    &raw_target_name, &raw_cred, &raw_ctx,
                                    &raw_mech_type, &flags_list, &ttl,
                                    &raw_channel_bindings, &raw_input_token,
                                    &raw_input_token_len))
        return NULL;

    gss_name_t target_name = GET_CAPSULE(gss_name_t, raw_target_name);

    gss_cred_id_t cred = CAPSULE_OR_DEFAULT(gss_cred_id_t, raw_cred,
                                            GSS_C_NO_CREDENTIAL);

    gss_ctx_id_t ctx = CAPSULE_OR_DEFAULT(gss_ctx_id_t, raw_ctx,
                                          GSS_C_NO_CONTEXT);

    gss_OID mech_type = CAPSULE_OR_DEFAULT(gss_OID, raw_mech_type,
                                           GSS_C_NO_OID);

    OM_uint32 req_flags = parseFlags(flags_list,
                                    (GSS_C_MUTUAL_FLAG | GSS_C_SEQUENCE_FLAG));

    gss_channel_bindings_t bdng = CAP_OR_DFLT_DR(gss_channel_bindings_t,
                                                 raw_channel_bindings,
                                                 GSS_C_NO_CHANNEL_BINDINGS);

    gss_buffer_desc input_token = GSS_C_EMPTY_BUFFER;

    if (raw_input_token && *raw_input_token)
    {
        input_token.length = raw_input_token_len;
        input_token.value = raw_input_token;
    }

    gss_OID actual_mech_type;
    gss_buffer_desc output_token = GSS_C_EMPTY_BUFFER;
    OM_uint32 ret_flags;
    OM_uint32 output_ttl;

    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    Py_BEGIN_ALLOW_THREADS
        maj_stat = gss_init_sec_context(&min_stat, cred, &ctx, target_name,
                                        mech_type, req_flags, ttl, bdng,
                                        &input_token, &actual_mech_type,
                                        &output_token, &ret_flags,
                                        &output_ttl);
    Py_END_ALLOW_THREADS

    if (maj_stat == GSS_S_COMPLETE || maj_stat == GSS_S_CONTINUE_NEEDED) {
        PyObject *cap_ctx = PyCapsule_New(ctx, NULL, NULL);
        PyObject *cap_mech_type = createMechType(self, actual_mech_type);
        PyObject *reqs_out = createFlagsList(self, ret_flags);
        PyObject *continue_needed;
        if (maj_stat == GSS_S_CONTINUE_NEEDED) {
            continue_needed = Py_True;
            Py_INCREF(Py_True);
        }
        else {
            continue_needed = Py_False;
            Py_INCREF(Py_False);
        }

        PyObject *res = Py_BuildValue(BYTE_TUPLE_STR("OOO", "IO"),
                                      cap_ctx, cap_mech_type,
                                      reqs_out, output_token.value,
                                      output_token.length, output_ttl,
                                      continue_needed);

        gss_release_buffer(&min_stat, &output_token);
        Py_DECREF(reqs_out);
        Py_DECREF(cap_ctx);
        Py_DECREF(cap_mech_type);
        Py_DECREF(continue_needed);
        return res;
    }
    else {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }
}

#define INT_OR_DEFAULT(obj, def) VALUE_OR_DEFAULT(obj, PyInt_AsLong(obj), def)

static PyObject *
getMIC(PyObject *self, PyObject *args)
{
    PyObject *raw_ctx;
    const char *message;
    int message_len;
    PyObject *raw_qop = Py_None;

    if (!PyArg_ParseTuple(args, "Os#|O", &raw_ctx, &message,
                          &message_len, &raw_qop))
        return NULL;

    gss_ctx_id_t ctx = GET_CAPSULE_DEREF(gss_ctx_id_t, raw_ctx);
    gss_buffer_desc message_buffer = GSS_C_EMPTY_BUFFER;
    OM_uint32 qop_req = INT_OR_DEFAULT(raw_qop, GSS_C_QOP_DEFAULT);

    message_buffer.length = message_len;
    message_buffer.value = message;

    gss_buffer_desc token_buffer = GSS_C_EMPTY_BUFFER;

    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    Py_BEGIN_ALLOW_THREADS
        maj_stat = gss_get_mic(&min_stat, ctx, qop_req, &message_buffer,
                               &token_buffer);
    Py_END_ALLOW_THREADS

    if (maj_stat == GSS_S_COMPLETE) {
        PyObject *res = Py_BuildValue(BYTE_STR, token_buffer.value,
                                                token_buffer.length);

        OM_uint32 release_min_stat;
        gss_release_buffer(&release_min_stat, &token_buffer);

        return res;
    }
    else {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }
}

static PyObject *
verifyMIC(PyObject *self, PyObject *args)
{
    PyObject *raw_ctx;
    const char *message;
    int message_len;
    const char *token;
    int token_len;
    int return_bool = 0;

    if (!PyArg_ParseTuple(args, "Os#s#|I", &raw_ctx, &message, &message_len,
                          &token, &token_len, &return_bool))
        return NULL;

    gss_ctx_id_t ctx = GET_CAPSULE_DEREF(gss_ctx_id_t, raw_ctx);
    gss_buffer_desc message_buffer = GSS_C_EMPTY_BUFFER;
    gss_buffer_desc token_buffer = GSS_C_EMPTY_BUFFER;

    message_buffer.length = message_len;
    message_buffer.value = message;

    token_buffer.length = token_len;
    token_buffer.value = token;

    gss_qop_t qop_state;

    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    Py_BEGIN_ALLOW_THREADS
        maj_stat = gss_verify_mic(&min_stat, ctx, &message_buffer,
                                  &token_buffer, &qop_state);
    Py_END_ALLOW_THREADS

    if (maj_stat == GSS_S_COMPLETE || maj_stat == GSS_S_DUPLICATE_TOKEN) {
        if (return_bool) {
            return Py_BuildValue("OIII", Py_True, qop_state,
                                 maj_stat, min_stat);
        }
        else {
            return PyInt_FromLong(qop_state);
        }
    }
    else {
        if (return_bool) {
            return Py_BuildValue("OIII", Py_False, qop_state,
                                 maj_stat, min_stat);
        }
        else {
            raise_gss_error(self, maj_stat, min_stat);
            return NULL;
        }
    }
}

static PyObject *
wrapSizeLimit(PyObject *self, PyObject *args, PyObject **keywds)
{
    PyObject *raw_ctx;
    int message_size;
    int conf_req = 1;
    int qop_req = 0;

    static char *kwlist[] = {"ctx", "output_size", "confidential", "qop"};

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "OI|II", kwlist,
                                      &raw_ctx, &message_size,
                                      &conf_req, &qop_req))
        return NULL;

    gss_ctx_id_t ctx = GET_CAPSULE_DEREF(gss_ctx_id_t, raw_ctx);

    OM_uint32 max_input_size;

    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    Py_BEGIN_ALLOW_THREADS
        maj_stat = gss_wrap_size_limit(&min_stat, ctx, conf_req, qop_req,
                                       message_size, &max_input_size);
    Py_END_ALLOW_THREADS

    if (maj_stat == GSS_S_COMPLETE) {
        return PyInt_FromLong(max_input_size);
    }
    else {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }
}

static PyObject *
wrap(PyObject *self, PyObject *args)
{
    PyObject *raw_ctx;
    const char *message;
    int message_len;
    int conf_req = 1;
    PyObject *raw_qop = Py_None;

    if(!PyArg_ParseTuple(args, "Os#|iO", &raw_ctx, &message, &message_len,
                         &conf_req, &raw_qop))
        return NULL;

    OM_uint32 qop_req = INT_OR_DEFAULT(raw_qop, GSS_C_QOP_DEFAULT);

    gss_buffer_desc input_message_buffer = GSS_C_EMPTY_BUFFER;

    gss_ctx_id_t ctx = GET_CAPSULE_DEREF(gss_ctx_id_t, raw_ctx);

    input_message_buffer.length = message_len;
    input_message_buffer.value = message;

    gss_buffer_desc output_message_buffer = GSS_C_EMPTY_BUFFER;
    int conf_state;
    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    Py_BEGIN_ALLOW_THREADS
        maj_stat = gss_wrap(&min_stat, ctx, conf_req, qop_req,
                            &input_message_buffer, &conf_state,
                            &output_message_buffer);
    Py_END_ALLOW_THREADS

    if (maj_stat == GSS_S_COMPLETE) {
        PyObject *conf_state_out;
        if (conf_state) {
            conf_state_out = Py_True;
            Py_INCREF(Py_True);
        }
        else {
            conf_state_out = Py_False;
            Py_INCREF(Py_False);
        }

        PyObject *res = Py_BuildValue(BYTE_TUPLE_STR("", "O"),
                                      output_message_buffer.value,
                                      output_message_buffer.length,
                                      conf_state_out);

        gss_release_buffer(&min_stat, &output_message_buffer);
        Py_DECREF(conf_state_out);
        return res;
    }
    else {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }
}

static PyObject *
unwrap(PyObject *self, PyObject *args)
{
    PyObject *raw_ctx;
    const char *message;
    int message_len;

    if(!PyArg_ParseTuple(args, "Os#", &raw_ctx, &message, &message_len))
        return NULL;

    gss_buffer_desc input_message_buffer = GSS_C_EMPTY_BUFFER;
    gss_ctx_id_t ctx = GET_CAPSULE_DEREF(gss_ctx_id_t, raw_ctx);

    input_message_buffer.length = message_len;
    input_message_buffer.value = message;

    gss_buffer_desc output_message_buffer = GSS_C_EMPTY_BUFFER;
    int conf_state;
    gss_qop_t qop_state;
    OM_uint32 maj_stat;
    OM_uint32 min_stat;

    Py_BEGIN_ALLOW_THREADS
        maj_stat = gss_unwrap(&min_stat, ctx, &input_message_buffer,
                              &output_message_buffer, &conf_state, &qop_state);
    Py_END_ALLOW_THREADS

    if (maj_stat == GSS_S_COMPLETE)
    {
        PyObject *conf_state_out;
        if (conf_state) {
            conf_state_out = Py_True;
            Py_INCREF(Py_True);
        }
        else {
            conf_state_out = Py_False;
            Py_INCREF(Py_False);
        }

        PyObject *res = Py_BuildValue(BYTE_TUPLE_STR("", "OI"),
                                      output_message_buffer.value,
                                      output_message_buffer.length,
                                      conf_state_out, qop_state);
        gss_release_buffer(&min_stat, &output_message_buffer);
        Py_DECREF(conf_state_out);
        return res;
    }
    else {
        raise_gss_error(self, maj_stat, min_stat);
        return NULL;
    }
}

static PyMethodDef GSSAPIMethods[] = {
    {"importName", importName, METH_VARARGS,
     "Convert a string name and type into a GSSAPI name object"},
    {"displayName", displayName, METH_O,
     "Convert a GSSAPI name back into a string and a NameType"},
    {"compareName", compareName, METH_VARARGS,
     "Compare Two GSSAPI Names"},
    {"exportName", exportName, METH_O,
     "Export a GSSAPI Mechanism Name"},
    {"canonicalizeName", canonicalizeName, METH_VARARGS,
     "Canonicalize an Abitrary GSSAPI Name into a Mechanim Name"},
    {"duplicateName", duplicateName, METH_O,
     "Duplicate a GSSAPI Name"},
    {"indicateMechs", indicateMechs, METH_NOARGS,
     "List the mechanisms supported by the current implementation"},
    {"acquireCred", acquireCred, METH_VARARGS | METH_KEYWORDS,
     "Acquire credentials from a name object"},
    {"releaseName", releaseName, METH_O,
     "Release a GSSAPI name"},
    {"releaseCred", releaseCred, METH_O,
     "Release GSSAPI credentials"},
    {"initSecContext", initSecContext, METH_VARARGS | METH_KEYWORDS,
     "Initialize a GSS security context"},
    {"acceptSecContext", acceptSecContext, METH_VARARGS | METH_KEYWORDS,
     "Accept a GSS security context"},
    {"deleteSecContext", deleteSecContext, METH_VARARGS,
     "Release a GSS security context"},
    {"getMIC", getMIC, METH_VARARGS,
     "Generate a cryptographic MIC for a message"},
    {"verifyMIC", verifyMIC, METH_VARARGS,
     "Verify that a MIC matches a message"},
    {"wrapSizeLimit", wrapSizeLimit, METH_VARARGS | METH_KEYWORDS,
     "Calculate the maximum input message size for an output message size"},
    {"unwrap", unwrap, METH_VARARGS,
     "Unwrap and possibly decrypt a message"},
    {"wrap", wrap, METH_VARARGS,
     "Wrap and possibly encrypt a message"},
    {"getMechanismType", getMechanismType, METH_VARARGS,
     "convert a value from the MechType enum into a mechanism type for use with GSSAPI methods"},
    {NULL, NULL, 0, NULL} /* sentinel value */
};

MOD_INIT(impl)
{
    PyObject *module;
    DEFINE_MODULE(module, "gssapi.base.impl", GSSAPIMethods,
                  impl_traverse, impl_clear, struct module_state);

    PyObject *types_module = PyImport_ImportModule("gssapi.base.types");

        PyObject *gsserror_attr_name = PyString_FromString("GSSError");

            GETSTATE(module)->GSSError_class =
                PyObject_GetAttr(types_module, gsserror_attr_name);

        Py_DECREF(gsserror_attr_name); /* we don't need it any more */

        PyObject *requirementflag_attr_name = PyString_FromString("RequirementFlag");

            GETSTATE(module)->RequirementFlag_class =
                PyObject_GetAttr(types_module, requirementflag_attr_name);

        Py_DECREF(requirementflag_attr_name); /* we don't need it any more */

        PyObject *mechtype_attr_name = PyString_FromString("MechType");

            GETSTATE(module)->MechType_class =
                PyObject_GetAttr(types_module, mechtype_attr_name);

        Py_DECREF(mechtype_attr_name); /* we don't need it any more */

        PyObject *nametype_attr_name = PyString_FromString("NameType");

            GETSTATE(module)->NameType_class =
                PyObject_GetAttr(types_module, nametype_attr_name);

        Py_DECREF(nametype_attr_name); /* we don't need it any more */

    Py_DECREF(types_module); /* we don't need it any more */

    return MOD_SUCCESS(module);
}
